//==========================================================================
//                 Header-Datei fuer das SP-System
//==========================================================================
#ifndef sp_h
#define sp_h
#define sp_test   // falls definiert: Testversion
#include<string.h>
#include<fstream.h>
#include<iomanip.h>
#include<stdlib.h>
#include<stdio.h>
#include<math.h>
#include<new.h>
#define Real double
#define Int unsigned int
#define eps4 1.e-6     // Genauigkeitsschranke bei Real = float
#define eps8 1.e-10    // Genauigkeitsschranke bei Real = double
#define dr4  15        // Druckspalten-Anzahl bei Real = float
#define dr8  25        // Druckspalten-Anzahl bei Real = double
#define More 256       // NNE-Anzahl bei Matrix-Erweiterung

// Methoden-Feld
static const char *sp_meth[] = {"MAT","LU","LUpiv","LUinc",\
                                "LDLT","LDLTpiv","LDLTinc"};
// Indices der Methoden
#define sp_MAT     0
#define sp_LU      1
#define sp_LUpiv   2
#define sp_LUinc   3
#define sp_LDLT    4
#define sp_LDLTpiv 5
#define sp_LDLTinc 6

#define sp_END    15

struct Mat_el          // externe Datenstruktur eines Matrixelementes
{ Int     i,           // Zeilenindex  (beginnt mit 1)
          j;           // Spaltenindex  (beginnt mit 1)
  Real    x;           // NNE
};

struct Vec_el  // externe Datenstruktur eines Vektorelementes
{ Int     i;   // Komponentenindex  (mit 1 beginnend)
  Real    x;   // Wert
};

struct NNE    // Datenstruktur eines NNE
{ Int     i,  // Zeilenindex  (beginnend mit 1)
          j;  // Spaltenindex  (beginnend mit 1)
  Real    x;  // NNE
  NNE    *z,  // Zeiger auf naechstes NNE in der Zeile oder auf NNE-Feld
         *s;  // Zeiger auf naechstes NNE in der Spalte oder auf NNE-Feld
};

/*
    NNE     *A; NNE-Feld
    A->z:       Zeiger auf das 1. NNE der Leerkette,
    A->i:       m+1,
    A->j:       n+1,
    A->s:       Zeiger auf den letzten Block
   fuer 1<=i<=m:
   (A+i)->i:   NNE-Anzahl der Zeile,
   (A+i)->j:   NNE-Anzahl der Spalte,
   (A+i)->x:   Skalierungsfaktor bzw. Diagonalelement
   (A+i)->z:   Zeiger auf das 1. NNE der i-ten Zeile oder auf A
   (A+i)->s:   Zeiger auf das 1. NNE der i-ten Spalte oder auf A
*/

void sp_error_(const char *, Int  = 0);
Real ausl(Real, Real);

class sp_Matrix; class sp_qMatrix;
class sp_sMatrix; class sp_bMatrix; class sp_Vector;

class ls_Org
{ protected:
  Int *copy;     // Kopie-Zaehler
public:
  Int asp;       // Anzahl der Druckspalten [double: dr8;  float: dr4]
  Real eps;      // Genauigkeitsschranke [double: eps8;  float: eps4]
  char *oname;   // Identifikator
  ls_Org(char * =NULL);
  ls_Org(const ls_Org &);
  ~ls_Org();
};

class ls_Vector: public ls_Org
{ friend sp_Matrix; friend sp_qMatrix;
  friend sp_sMatrix; friend sp_bMatrix;  friend sp_Vector;
  friend ls_Vector operator*(Real, const ls_Vector &);
  Real *A;
  Int  n;
  public:
  ls_Vector(Int, char* = NULL);
  Real& operator[] (Int);                  // Komponenten-Ein/Ausgabe
  ~ls_Vector();
  Int dimension(){ return n; }             // Dimension
  void write(ostream &);
  void read(istream &);
  void operator>> (char *);
  void operator<< (char *);
  Int operator== (const ls_Vector &) const;   // Objekt1 == Objekt2
  Int operator != (const ls_Vector &) const;  // Objekt1  != Objekt2
  ls_Vector operator- () const;               // - Objekt
  ls_Vector operator= (Real);                 // alle Komponenten ein Wert
  ls_Vector operator* (Real) const;           // Objekt mal Wert
  ls_Vector operator*= (Real);                // Objekt *= Objekt
  void operator= (const ls_Vector &);         // Objekt1 = Objekt2
  ls_Vector operator= (const sp_Vector &);
  ls_Vector operator+ (const ls_Vector &) const; // Objekt1 + Objekt2
  ls_Vector operator- (const ls_Vector &) const; // Objekt1 - Objekt2
  ls_Vector operator += (const ls_Vector &);     // Objekt += Objekt
  ls_Vector operator-= (const ls_Vector &);      // Objekt -= Objekt
  Real operator* (const ls_Vector &) const;   // Skalarprodukt
};

class sp_Org
{ protected:
  Int *copy;
  void sp_zero_(NNE **, Int *, Int = 0, Int = 0);  // Block-Zuteilung
  void sp_delete_(NNE *);   // Block-Rueckgabe
  public:
  Int asp,       // Anzahl der Druckspalten [double: dr8;  float: dr4]
      *rc;       // sekundaerer Rueckkehrcode bei manchen Funktionen
                 //   ( z. B. Iterationszahl bei iterativen Verfahren)
  Real eps;      // Genauigkeitsschranke [double: eps8;  float: eps4]
  char *oname;   // Identifikator
  sp_Org(char * = NULL);
  sp_Org(const sp_Org &);
  ~sp_Org();
};

class sp_Vector: public sp_Org
{ friend sp_Matrix; friend sp_qMatrix;
  friend sp_sMatrix; friend ls_Vector;
  friend sp_Vector operator*(Real, const sp_Vector &);
  Vec_el *A;
  Int   n,
      n_A,     // NNE-Anzahl
      l_A;     // Feldlaenge von A
  public:
  sp_Vector(Int, Int = 0, char* = NULL);        // Konstruktor
  Real operator[] (Int);                       // Komponenten-Ausgabe
  ~sp_Vector();
  Int dimension(){ return n; }                 // Dimension
  void write(ostream &);
  void read(istream &);
  void operator>> (char *);
  void operator<< (char *);
  void operator<< (const Vec_el &);      // aufsteigende Komponenten-Eingabe
  Int operator== (const sp_Vector &) const;     // Objekt1 == Objekt2
  Int operator!= (const sp_Vector &) const;     // Objekt1  != Objekt2
  sp_Vector operator- () const;                 // - Objekt
  sp_Vector operator* (Real) const;             // Objekt mal Wert
  void operator*= (Real);                      // Objekt *= Objekt
  void operator= (const sp_Vector &);           // Objekt1 = Objekt2
  void operator= (const ls_Vector &);
  sp_Vector operator+ (const sp_Vector &) const; // Objekt1 + Objekt2
  sp_Vector operator- (const sp_Vector &) const; // Objekt1 - Objekt2
  void operator+= (const sp_Vector &);          // Objekt += Objekt
  void operator-= (const sp_Vector &);          // Objekt -= Objekt
  Real operator* (const sp_Vector &) const;     // Skalarprodukt
};

// Methoden-Maske der Klasse sp_Matrix
static const unsigned char sp_m_mask[] = {sp_MAT, sp_END};


class sp_Matrix: public sp_Org                  // RechteckMatrix
{ friend sp_Vector; friend sp_qMatrix;
  friend sp_Matrix operator* (Real, const sp_Matrix &);
  protected:
  NNE   *A;     // NNE-Feld
  Int    m,     // Zeilenzahl
         n,     // Spaltenzahl  (n<=m)
       l_A;     // Anzahl der freien Speicherplaetze
  unsigned char typ;
  virtual int get_name(){ return -1;}
  void change_row(Int, Int);
  void change_column(Int, Int);
  public:
  sp_Matrix(Int, Int, char* = "MAT", char* = NULL);
  ~sp_Matrix();
  Int row_dimension() { return m; }
  Int column_dimension() { return n; }
  Real filling_density();
  sp_Matrix operator* (const sp_Matrix &) const;
  virtual void write(ostream &);
  virtual void read(istream &);
  virtual void operator>> (char *);
  virtual void operator<< (char *);
  virtual void operator<< (const Mat_el &);          // NNE-Eingabe
  virtual Real operator>> (Mat_el *);                // NNE-Ausgabe
  void print();
  void put_row(Int, const sp_Vector &);
  sp_Vector get_row(Int);
  void put_column(Int, const sp_Vector &);
  sp_Vector get_column(Int);
  void put_lower_diagonal(Int, const sp_Vector &);
  void put_lower_diagonal(Int, Real);
  sp_Vector get_lower_diagonal(Int);
  void put_upper_diagonal(Int, const sp_Vector &);
  void put_upper_diagonal(Int, Real);
  sp_Vector get_upper_diagonal(Int);
  void delete_nne(const Mat_el &);
  void delete_column(Int);
  void delete_in_column(Int, const sp_Vector &);
  void delete_row(Int);
  void delete_in_row(Int, const sp_Vector &);
  void delete_lower_diagonal(Int);
  void delete_in_lower_diagonal(Int, const sp_Vector &);
  void delete_upper_diagonal(Int);
  void delete_in_upper_diagonal(Int, const sp_Vector &);
  sp_Vector operator* (sp_Vector &) const;
  ls_Vector operator* (ls_Vector &) const;        // Matrix mal ls_Vector
  ls_Vector operator% (ls_Vector &) const;        // transpon. Matrix mal Vektor
  ls_Vector solve_with_MAT(ls_Vector &, ls_Vector &);// Loesen Ausgleichs-Problem
  ls_Vector solve(ls_Vector &);
};

// Methoden-Maske der Klasse sp_qMatrix
static const unsigned char sp_q_mask[] ={sp_MAT,sp_LU,sp_LUpiv,sp_LUinc,sp_END};

class sp_qMatrix: public sp_Matrix           // quadratisch
{ friend sp_Vector;
  protected:
  Int *icol, // Indexfeld fuer Spaltentausch
      *irow; // Indexfeld fuer Zeilentausch
  char *a_name,  // NULL oder Name der Auslagerungsdatei
                 // fuer die sparse-Matrix
       *u_name,  // NULL oder Name der Auslagerungsdatei
                 // fuer unvollstaendige Faktorisierung
       a_n, u_n;
  virtual int get_name(){ return -2;}
  public:
  sp_qMatrix(Int, char* = "MAT", char* = NULL);
  ~sp_qMatrix();
  Int dimension(){ return n; }
  virtual void write(ostream &);
  virtual void read(istream &);
  virtual void operator>> (char *);
  virtual void operator<< (char *);
  virtual void operator<< (const Mat_el &);          // NNE-Eingabe
  virtual Real operator>> (Mat_el *);                // NNE-Ausgabe
  ls_Vector solve_with_LU(ls_Vector &);
  void LU(){ ls_Vector b(0); solve_with_LU(b); }
  ls_Vector solve_with_LUpiv(ls_Vector &);
  void LUpiv(){ ls_Vector b(0); solve_with_LUpiv(b); }
  ls_Vector solve_with_LUinc(ls_Vector &, ls_Vector &);
  void LUinc(){ ls_Vector b(0); solve_with_LUinc(b,b); }
  ls_Vector solve(ls_Vector &);
};

// Methoden-Maske der Klasse sp_sMatrix
static const unsigned char \
   sp_s_mask[] = {sp_MAT,sp_LDLT,sp_LDLTpiv,sp_LDLTinc,sp_END};

class sp_sMatrix: public sp_Matrix             // symmetrisch
{ friend sp_Vector;
  friend sp_sMatrix operator* (Real, const sp_sMatrix &);
  protected:
  Int *icol; // Indexfeld fuer Zeile/Spaltentausch
  char *a_name,  // NULL oder Name der Auslagerungsdatei
                 // fuer die sparse-Matrix
       *u_name,  // NULL oder Name der Auslagerungsdatei
                 // fuer unvollstaendige Faktorisierung
       a_n, u_n;
  virtual int get_name(){ return 0;}
  void change_diagonal(Int, Int);
  public:
  sp_sMatrix(Int, char* = "MAT", char* = NULL);
  ~sp_sMatrix();
  Int dimension(){ return n; }
  virtual void write(ostream &);
  virtual void read(istream &);
  virtual void operator>> (char *);
  virtual void operator<< (char *);
  virtual void operator<< (const Mat_el &);          // NNE-Eingabe
  virtual Real operator>> (Mat_el *);                // NNE-Ausgabe
  sp_Vector operator* (sp_Vector &) const; // Matrix mal Vektor
  ls_Vector operator* (ls_Vector &) const;     // Matrix mal Vektor
  ls_Vector solve_with_MAT(ls_Vector &, ls_Vector &);
  ls_Vector solve_with_LDLT(ls_Vector &);
  void LDLT(){ ls_Vector b(0); solve_with_LDLT(b); }
  ls_Vector solve_with_LDLTpiv(ls_Vector &);
  void LDLT_piv(){ ls_Vector b(0); solve_with_LDLTpiv(b); }
  ls_Vector solve_with_LDLTinc(ls_Vector &, ls_Vector &);
  void LDLTinc(){ ls_Vector b(0); solve_with_LDLTinc(b, b); }
  ls_Vector solve(ls_Vector &);
};

// Methoden-Maske der Klasse sp_bMatrix
static const unsigned char sp_b_mask[] = {sp_MAT, sp_END};

class sp_bMatrix: public sp_Org   // symmetrische Band-Matrix
{ unsigned char *o_f, typ;
  Int n,
       o_anz,// Anzahl der oberen Nebendiagonalen (ohne Hauptdiagonale)
      *o_ind;// o_ind[i] Index der i-ten oberen Nebendiagonalen
  Real **o_w;// *o_w[i]: Wertefeld der oberen Nebendiagonalen
  public:
  sp_bMatrix(Int, Int, char* = "MAT", char* = NULL);
  ~sp_bMatrix();
  Int dimension() { return n; }
  Int number_of_diagonals() { return o_anz; }
  void write(ostream &);
  void read(istream &);
  void operator>> (char *);
  void operator<< (char *);
  void operator<< (const Mat_el &);          // NNE-Eingabe
  Real operator>> (Mat_el *);                // NNE-Ausgabe
  void put_upper_diagonal(Int, const ls_Vector &);
  void put_upper_diagonal(Int, Real);
  ls_Vector get_upper_diagonal(Int);
  void delete_upper_diagonal(Int);
  ls_Vector operator* (ls_Vector &) const;         // Matrix mal Vektor
  ls_Vector solve_with_MAT(ls_Vector &, ls_Vector &);
  ls_Vector solve(ls_Vector &);
};

ls_Vector operator* (Real, const ls_Vector &);     // Wert mal Vektor
#endif

